"""Subclass specific to Cisco ASA"""

from aionet.vendors.devices.base_ios import BaseIOSDevice


class CiscoASA(BaseIOSDevice):
    """Class for working with Cisco ASA"""

    def __init__(self, *args, **kwargs):
        """
        Initialize class for asynchronous working with network devices

        :param str host: device hostname or ip address for connection
        :param str username: username for logging to device
        :param str password: user password for logging to device
        :param str secret: secret password for privilege mode
        :param int port: ssh port for connection. Default is 22
        :param str device_type: network device type
        :param known_hosts: file with known hosts. Default is None (no policy). With () it will use default file
        :param str local_addr: local address for binding source of tcp connection
        :param client_keys: path for client keys. Default in None. With () it will use default file in OS
        :param str passphrase: password for encrypted client keys
        :param float timeout: timeout in second for getting information from channel
        :param loop: asyncio loop object
        """
        super().__init__(*args, **kwargs)
        self._multiple_mode = False

    _disable_paging_command = "terminal pager 0"

    @property
    def multiple_mode(self):
        """ Returning Bool True if ASA in multiple mode"""
        return self._multiple_mode

    async def _session_preparation(self):
        await super()._session_preparation()
        await self._check_multiple_mode()

    async def _disable_width(self):
        self._logger.info("setting terminal width to 511")
        await self.send_config_set([type(self)._disable_width_command])

    async def _check_multiple_mode(self):
        """Check mode multiple. If mode is multiple we adding info about contexts"""
        self._logger.info("Checking multiple mode")
        out = await self.send_command_expect("show mode")
        if "multiple" in out:
            self._multiple_mode = True

        self._logger.debug("Multiple mode: %s" % self._multiple_mode)



"""
Constants Module
"""
# Session Terminal Const.
TERM_WID = 2147483647
TERM_LEN = 2147483647
TERM_TYPE = 'vt100'

# ansi codes
CODE_SAVE_CURSOR = chr(27) + r"7"
CODE_SCROLL_SCREEN = chr(27) + r"\[r"
CODE_RESTORE_CURSOR = chr(27) + r"8"
CODE_CURSOR_UP = chr(27) + r"\[\d+A"
CODE_CURSOR_DOWN = chr(27) + r"\[\d+B"

CODE_POSITION_CURSOR = chr(27) + r"\[\d+;\d+H"
CODE_SHOW_CURSOR = chr(27) + r"\[\?25h"
CODE_NEXT_LINE = chr(27) + r"E"
CODE_ERASE_LINE = chr(27) + r"\[2K"
CODE_ENABLE_SCROLL = chr(27) + r"\[\d+;\d+r"

CODE_SET = [
    CODE_SAVE_CURSOR,
    CODE_SCROLL_SCREEN,
    CODE_RESTORE_CURSOR,
    CODE_CURSOR_UP,
    CODE_CURSOR_DOWN,
    CODE_POSITION_CURSOR,
    CODE_SHOW_CURSOR,
    CODE_ERASE_LINE,
    CODE_ENABLE_SCROLL,
]


from aionet.vendors.terminal_modes.juniper import CliMode
from aionet.vendors.devices.base_junos import BAseJunOSDevice


class JuniperJunOS(BAseJunOSDevice):
    """Class for working with Juniper JunOS"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.cli_mode = CliMode(
            enter_command=type(self)._cli_command,
            check_string=type(self)._cli_check,
            exit_command='',
            device=self
        )

    _cli_check = ">"
    """Checking string for shell mode"""

    _cli_command = "cli"
    """Command for entering to cli mode"""

    async def _session_preparation(self):
        await self.cli_mode()
        await super()._session_preparation()



"""
Device Dispatcher

"""
from aionet.vendors.devices import AristaEOS
from aionet.vendors.devices import ArubaAOS6, ArubaAOS8
from aionet.vendors.devices import CiscoASA, CiscoIOS, CiscoIOSXR, CiscoNXOS
from aionet.vendors.devices import FujitsuSwitch
from aionet.vendors.devices import HPComware, HPComwareLimited
from aionet.vendors.devices import JuniperJunOS
from aionet.vendors.devices import MikrotikRouterOS
from aionet.vendors.devices import Terminal
from aionet.vendors.devices import UbiquityEdgeSwitch


DEVICE_MAPPER = {
    "arista_eos": AristaEOS,
    "aruba_aos_6": ArubaAOS6,
    "aruba_aos_8": ArubaAOS8,
    "cisco_asa": CiscoASA,
    "cisco_ios": CiscoIOS,
    "cisco_ios_xe": CiscoIOS,
    "cisco_ios_xr": CiscoIOSXR,
    "cisco_nxos": CiscoNXOS,
    "fujitsu_switch": FujitsuSwitch,
    "hp_comware": HPComware,
    "hp_comware_limited": HPComwareLimited,
    "juniper_junos": JuniperJunOS,
    "mikrotik_routeros": MikrotikRouterOS,
    "ubiquity_edge": UbiquityEdgeSwitch,
    "terminal": Terminal,
}


platforms = list(DEVICE_MAPPER.keys())
platforms.sort()
platforms_str = u"\n".join(platforms)


def ConnectionHandler(*args, **kwargs):
    if kwargs["device_type"] not in platforms:
        raise ValueError(
            "Unsupported device_type: "
            "currently supported platforms are: {0}".format(platforms_str)
        )
    connection_class = DEVICE_MAPPER[kwargs["device_type"]]
    return connection_class(*args, **kwargs)



from aionet.vendors.terminal_modes.hp import CmdLineMode
from aionet.vendors.devices.base_comware import BaseComwareDevice


class HPComwareLimited(BaseComwareDevice):
    """Class for working with HP Comware Limited like 1910 and 1920 models"""

    def __init__(self, cmdline_password=u"", *args, **kwargs):
        """
        Initialize  class for asynchronous working with network devices

        :param str host: device hostname or ip address for connection
        :param str username: username for logging to device
        :param str password: user password for logging to device
        :param str cmdline_password: password for entering to _cmd_line
        :param int port: ssh port for connection. Default is 22
        :param str device_type: network device type
        :param known_hosts: file with known hosts. Default is None (no policy). With () it will use default file
        :param str local_addr: local address for binding source of tcp connection
        :param client_keys: path for client keys. Default in None. With () it will use default file in OS
        :param str passphrase: password for encrypted client keys
        :param float timeout: timeout in second for getting information from channel
        :param loop: asyncio loop object
        """
        super().__init__(*args, **kwargs)
        self.cmdline_mode = CmdLineMode(
            enter_command=type(self)._cmdline_mode_enter_command,
            check_error_string=type(self)._cmdline_mode_check,
            password=cmdline_password,
            device=self
        )

    _cmdline_mode_enter_command = "_cmdline-mode on"
    """Command for entering to cmdline model"""

    _cmdline_mode_check = "Invalid password"
    """Checking string for wrong password in trying of entering to cmdline mode"""

    async def _session_preparation(self):
        await self.cmdline_mode()
        await super()._session_preparation()



"""
Hp Terminal Modes
"""
from aionet.logging import logger
from aionet.vendors.terminal_modes.base import BaseTerminalMode


class SystemView(BaseTerminalMode):
    """ System View Terminal mode """
    name = 'system_view'
    pass


class CmdLineMode:
    """ CmdLine Terminal Mode """
    _name = 'cmdline'

    def __init__(self,
                 enter_command,
                 check_error_string,
                 password,
                 device):
        self._enter_command = enter_command
        self._check_error_string = check_error_string
        self._password = password
        self.device = device

    def __call__(self, *args, **kwargs):
        return self.enter()

    @property
    def _logger(self):
        return self.device._logger

    async def enter(self):
        """Entering to cmdline-mode"""
        self._logger.info("Entering to cmdline mode")

        output = await self.device.send_command(self._enter_command, pattern="\[Y\/N\]")
        output += await self.device.send_command("Y", pattern="password\:")
        output += await self.device.send_command(self._password)

        logger.debug("cmdline mode output: %s" % repr(output))

        logger.info("Checking cmdline mode")
        if self._check_error_string in output:
            raise ValueError("Failed to enter to cmdline mode")
        self.device.current_terminal = self

        return output



"""
Telnet Connection Module
"""
import asyncio
from aionet.exceptions import AionetAuthenticationError, AionetTimeoutError
from aionet.connections.base import BaseConnection


class TelnetConnection(BaseConnection):
    def __init__(self,
                 host=u"",
                 username=u"",
                 password=u"",
                 port=23,
                 timeout=15,
                 loop=None,
                 pattern=None, ):
        super().__init__()
        if host:
            self._host = host
        else:
            raise ValueError("Host must be set")
        self._port = int(port)
        self._timeout = timeout
        self._username = username
        self._password = password
        if loop is None:
            self._loop = asyncio.get_event_loop()
        else:
            self._loop = loop

        if pattern is not None:
            self._pattern = pattern

        self._timeout = timeout

    async def _start_session(self):
        """ start Telnet Session by login to device """
        self._logger.info("Host {}: telnet: trying to login to device".format(self._host))
        output = await self.read_until_pattern(['username', 'Username'])
        self.send(self._username + '\n')
        output += await self.read_until_pattern(['password', 'Password'])
        self.send(self._password + '\n')
        output += await self.read_until_prompt()
        self.send('\n')
        if 'Login invalid' in output:
            raise AionetAuthenticationError(self._host, None, "authentication failed")

    def __check_session(self):
        if not self._stdin:
            raise RuntimeError("telnet session not started")

    @asyncio.coroutine
    def connect(self):
        """ Establish Telnet Connection """
        self._logger.info("Host {}: telnet: Establishing Telnet Connection on port {}".format(self._host, self._port))
        fut = asyncio.open_connection(self._host, self._port, family=0, flags=0)
        try:
            self._stdout, self._stdin = yield from asyncio.wait_for(fut, self._timeout)
        except asyncio.TimeoutError:
            raise AionetTimeoutError(self._host)
        except Exception as e:
            raise AionetAuthenticationError(self._host, None, str(e))

        yield from self._start_session()

    async def disconnect(self):
        """ Gracefully close the Telnet connection """
        self._logger.info("Host {}: telnet: Disconnecting".format(self._host))
        self._logger.info("Host {}: telnet: Disconnecting".format(self._host))
        self._conn.close()
        await self._conn.wait_closed()

    def send(self, cmd):
        self._stdin.write(cmd.encode())

    async def read(self):
        output = await self._stdout.read(self._MAX_BUFFER)
        return output.decode(errors='ignore')

    async def close(self):
        pass



"""
Juniper Terminal Modes
"""
from aionet.vendors.terminal_modes.base import BaseTerminalMode
from aionet.vendors.terminal_modes.cisco import ConfigMode as CiscoConfigMode


class ConfigMode(CiscoConfigMode):
    pass


class CliMode(BaseTerminalMode):
    name = 'cli_mode'

    def exit(self):
        pass



"""
SSH Connection Module
"""
import asyncio
import asyncssh
from aionet.constants import TERM_LEN, TERM_WID, TERM_TYPE
from aionet.exceptions import AionetAuthenticationError, AionetTimeoutError
from aionet.connections.base import BaseConnection


class SSHConnection(BaseConnection):
    def __init__(self,
                 host=u"",
                 username=u"",
                 password=u"",
                 port=22,
                 timeout=30,
                 loop=None,
                 known_hosts=None,
                 local_addr=None,
                 client_keys=None,
                 passphrase=None,
                 tunnel=None,
                 pattern=None,
                 agent_forwarding=False,
                 agent_path=(),
                 client_version=u"aionet-{}",
                 family=0,
                 kex_algs=(),
                 encryption_algs=(),
                 mac_algs=(),
                 compression_algs=(),
                 signature_algs=()):
        super().__init__()
        if host:
            self._host = host
        else:
            raise ValueError("Host must be set")
        self._port = int(port)
        self._timeout = timeout
        if loop is None:
            self._loop = asyncio.get_event_loop()
        else:
            self._loop = loop

        connect_params_dict = {
            "host": self._host,
            "port": self._port,
            "username": username,
            "password": password,
            "known_hosts": known_hosts,
            "local_addr": local_addr,
            "client_keys": client_keys,
            "passphrase": passphrase,
            "tunnel": tunnel,
            "agent_forwarding": agent_forwarding,
            "loop": loop,
            "family": family,
            "agent_path": agent_path,
            "client_version": client_version,
            "kex_algs": kex_algs,
            "encryption_algs": encryption_algs,
            "mac_algs": mac_algs,
            "compression_algs": compression_algs,
            "signature_algs": signature_algs
        }

        if pattern is not None:
            self._pattern = pattern

        self._conn_dict = connect_params_dict
        self._timeout = timeout

    async def connect(self):
        """ Etablish SSH connection """
        self._logger.info("Host {}: SSH: Establishing SSH connection on port {}".format(self._host, self._port))

        fut = asyncssh.connect(**self._conn_dict)
        try:
            self._conn = await asyncio.wait_for(fut, self._timeout)
        except asyncssh.DisconnectError as e:
            raise AionetAuthenticationError(self._host, e.code, e.reason)
        except asyncio.TimeoutError:
            raise AionetTimeoutError(self._host, None, 'timeout while connecting to %r' % self._host)

        await self._start_session()

    async def disconnect(self):
        """ Gracefully close the SSH connection """
        self._logger.info("Host {}: SSH: Disconnecting".format(self._host))
        self._logger.info("Host {}: SSH: Disconnecting".format(self._host))
        await self._cleanup()
        self._conn.close()
        await self._conn.wait_closed()

    def send(self, cmd):
        self._stdin.write(cmd)

    async def read(self):
        return await self._stdout.read(self._MAX_BUFFER)

    def __check_session(self):
        """ check session was opened """
        if not self._stdin:
            raise RuntimeError("SSH session not started")

    async def _start_session(self):
        """ start interactive-session (shell) """
        self._logger.info(
            "Host {}: SSH: Starting Interacive session term_type={}, term_width={}, term_length={}".format(
                self._host, TERM_TYPE, TERM_WID, TERM_LEN))
        self._stdin, self._stdout, self._stderr = await self._conn.open_session(
            term_type=TERM_TYPE, term_size=(TERM_WID, TERM_LEN)
        )

    async def _cleanup(self):
        pass

    async def close(self):
        """ Close Connection """
        await self._cleanup()
        self._conn.close()
        await self._conn.wait_closed()



"""
Base Connection Module
"""
import re
import asyncio
from aionet.logging import logger
from aionet.connections.interface import IConnection


class BaseConnection(IConnection):

    def __init__(self, *args, **kwargs):
        self._host = None
        self._timeout = None
        self._conn = None
        self._base_prompt = self._base_pattern = ""
        self._MAX_BUFFER = 65535

    async def __aenter__(self):
        """Async Context Manager"""
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Async Context Manager"""
        await self.disconnect()

    @property
    def _logger(self):
        return logger

    def set_base_prompt(self, prompt):
        """ base prompt setter """
        self._base_prompt = prompt

    def set_base_pattern(self, pattern):
        """ base patter setter """
        self._base_pattern = pattern

    async def disconnect(self):
        """ Close Connection """
        raise NotImplementedError("Connection must implement disconnect method")

    async def connect(self):
        """ Establish Connection """
        raise NotImplementedError("Connection must implement connect method")

    def send(self, cmd):
        """ send data """
        raise NotImplementedError("Connection must implement send method")

    async def read(self):
        """ read from buffer """
        raise NotImplementedError("Connection must implement read method ")

    async def read_until_pattern(self, pattern, re_flags=0, read_for=0):
        """Read channel until pattern detected. Return ALL data available"""

        if pattern is None:
            raise ValueError("pattern cannot be None")

        if isinstance(pattern, str):
            pattern = [pattern]
        output = ""
        logger.info("Host {}: Reading until pattern".format(self._host))

        logger.debug("Host {}: Reading pattern: {}".format(self._host, pattern))
        while True:

            fut = self.read()
            try:
                output += await asyncio.wait_for(fut, read_for or self._timeout)
            except asyncio.TimeoutError:
                if read_for:
                    return output
                raise TimeoutError(self._host)

            for exp in pattern:
                if re.search(exp, output, flags=re_flags):
                    logger.debug(
                        "Host {}: Reading pattern '{}' was found: {}".format(
                            self._host, pattern, repr(output)
                        )
                    )
                    return output

    async def read_until_prompt(self, read_for=0):
        """ read util prompt """
        return await self.read_until_pattern(self._base_pattern, read_for=read_for)

    async def read_until_prompt_or_pattern(self, pattern, re_flags=0, read_for=0):
        """ read util prompt or pattern """

        logger.info("Host {}: Reading until prompt or pattern".format(self._host))

        if isinstance(pattern, str):
            pattern = [self._base_prompt, pattern]
        elif isinstance(pattern, list):
            pattern = [self._base_prompt] + pattern
        else:
            raise ValueError("pattern must be string or list of strings")
        return await self.read_until_pattern(pattern=pattern, re_flags=re_flags, read_for=read_for)



"""
Utilities Module.
"""
import re, os
from clitable import CliTable, CliTableError
from aionet.constants import CODE_SET, CODE_NEXT_LINE


def strip_ansi_escape_codes(string):
    """
    Remove some ANSI ESC codes from the output

    http://en.wikipedia.org/wiki/ANSI_escape_code

    Note: this does not capture ALL possible ANSI Escape Codes only the ones
    I have encountered

    Current codes that are filtered:
    ESC = '\x1b' or chr(27)
    ESC = is the escape character [^ in hex ('\x1b')
    ESC[24;27H   Position cursor
    ESC[?25h     Show the cursor
    ESC[E        Next line (HP does ESC-E)
    ESC[2K       Erase line
    ESC[1;24r    Enable scrolling from start to row end
    ESC7         Save cursor position
    ESC[r        Scroll all screen
    ESC8         Restore cursor position
    ESC[nA       Move cursor up to n cells
    ESC[nB       Move cursor down to n cells

    require:
        HP ProCurve
        F5 LTM's
        Mikrotik
    """

    output = string
    for ansi_esc_code in CODE_SET:
        output = re.sub(ansi_esc_code, "", output)

    # CODE_NEXT_LINE must substitute with '\n'
    output = re.sub(CODE_NEXT_LINE, "\n", output)

    return output


def get_template_dir():
    """Find and return the ntc-templates/templates dir."""
    try:
        template_dir = os.environ["NET_TEXTFSM"]
        index = os.path.join(template_dir, "index")
        if not os.path.isfile(index):
            # Assume only base ./ntc-templates specified
            template_dir = os.path.join(template_dir, "templates")
    except KeyError:
        # Construct path ~/ntc-templates/templates
        home_dir = os.path.expanduser("~")
        template_dir = os.path.join(home_dir, "ntc-templates", "templates")

    index = os.path.join(template_dir, "index")
    if not os.path.isdir(template_dir) or not os.path.isfile(index):
        msg = """
Valid ntc-templates not found, please install https://github.com/networktocode/ntc-templates
and then set the NET_TEXTFSM environment variable to point to the ./ntc-templates/templates
directory."""
        raise ValueError(msg)
    return template_dir


def clitable_to_dict(cli_table):
    """Converts TextFSM cli_table object to list of dictionaries."""
    objs = []
    for row in cli_table:
        temp_dict = {}
        for index, element in enumerate(row):
            temp_dict[cli_table.header[index].lower()] = element
        objs.append(temp_dict)
    return objs


def get_structured_data(raw_output, platform, command):
    """Convert raw CLI output to structured data using TextFSM template."""
    template_dir = get_template_dir()
    index_file = os.path.join(template_dir, "index")
    textfsm_obj = CliTable(index_file, template_dir)
    attrs = {"Command": command, "Platform": platform}
    try:
        # Parse output through template
        textfsm_obj.ParseCmd(raw_output, attrs)
        structured_data = clitable_to_dict(textfsm_obj)
        output = raw_output if structured_data == [] else structured_data
        return output
    except CliTableError:
        return raw_output



from aionet.exceptions import AionetCommitError
from aionet.vendors.terminal_modes.cisco import IOSxrConfigMode
from aionet.vendors.devices.base_ios import BaseIOSDevice


class CiscoIOSXR(BaseIOSDevice):
    """Class for working with Cisco IOS XR"""

    _commit_command = "commit"
    """Command for committing changes"""

    _commit_comment_command = "commit comment {}"
    """Command for committing changes with comment"""

    _abort_command = "abort"
    """Command for aborting all changes and exit to privilege mode"""

    _show_config_failed = "show configuration failed"
    """Command for showing the reason of failed commit"""

    _show_commit_changes = "show configuration commit changes"
    """Command for showing the other commit which have occurred during our session"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.config_mode = IOSxrConfigMode(
            enter_command=type(self)._config_enter,
            exit_command=type(self)._config_check,
            check_string=type(self)._config_exit,
            device=self,
            parent=self.enable_mode
        )

    async def send_config_set(
            self,
            config_commands=None,
            with_commit=True,
            commit_comment="",
            exit_config_mode=True,
    ):
        """
        Sending configuration commands to device
        By default automatically exits/enters configuration mode.

        :param list config_commands: iterable string list with commands for applying to network devices in system view
        :param bool with_commit: if true it commit all changes after applying all config_commands
        :param string commit_comment: message for configuration commit
        :param bool exit_config_mode: If true it will quit from configuration mode automatically
        :return: The output of these commands
        """

        if config_commands is None:
            return ""

        # Send config commands
        output = await self.config_mode()
        output += await super(BaseIOSDevice, self).send_config_set(
            config_commands=config_commands
        )
        if with_commit:
            commit = type(self)._commit_command
            if commit_comment:
                commit = type(self)._commit_comment_command.format(commit_comment)

            output += await self.send_command_expect(
                commit,
                pattern=r"Do you wish to proceed with this commit anyway\?"
            )
            if "Failed to commit" in output:
                show_config_failed = type(self)._show_config_failed
                reason = await self.send_command_expect(show_config_failed)
                raise AionetCommitError(self.host, reason)
            if "One or more commits have occurred" in output:
                show_commit_changes = type(self)._show_commit_changes
                await self.send_command_expect('no')
                reason = await self.send_command_expect(show_commit_changes)
                raise AionetCommitError(self.host, reason)

        if exit_config_mode:
            output += await self.config_mode.exit()

        self._logger.debug("config commands output: %s" % repr(output))

        return output

    async def _cleanup(self):
        """ Any needed cleanup before closing connection """
        abort = type(self)._abort_command
        await self.send_command_expect(abort)
        self._logger.info("Cleanup session")



"""
Connection Interface
"""
import abc


class IConnection(abc.ABC):

    @abc.abstractmethod
    async def __aenter__(self):
        """Async Context Manager"""
        pass

    @abc.abstractmethod
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Async Context Manager"""
        pass

    @abc.abstractmethod
    async def disconnect(self):
        """ Close Connection """
        pass

    @abc.abstractmethod
    async def connect(self):
        """ Establish Connection """
        pass

    @abc.abstractmethod
    async def send(self, cmd):
        """ send Command """
        pass

    @abc.abstractmethod
    async def read(self):
        """ send Command """
        pass

    @abc.abstractmethod
    async def read_until_pattern(self, pattern, re_flags=0):
        """ read util pattern """
        pass

    @abc.abstractmethod
    async def read_until_prompt(self):
        """ read util pattern """
        pass

    @abc.abstractmethod
    async def read_until_prompt_or_pattern(self, attern, re_flags=0):
        """ read util pattern """
        pass



from aionet.vendors.devices.base import BaseDevice


class MikrotikRouterOS(BaseDevice):
    """Class for working with Mikrotik RouterOS"""

    def __init__(self, *args, **kwargs):
        """
        Initialize class for asynchronous working with network devices
        Invoke init with some special params (base_pattern and username)

        :param str host: device hostname or ip address for connection
        :param str username: username for logging to device
        :param str password: user password for logging to device
        :param int port: ssh port for connection. Default is 22
        :param str device_type: network device type
        :param known_hosts: file with known hosts. Default is None (no policy). With () it will use default file
        :param str local_addr: local address for binding source of tcp connection
        :param client_keys: path for client keys. Default in None. With () it will use default file in OS
        :param str passphrase: password for encrypted client keys
        :param float timeout: timeout in second for getting information from channel
        :param loop: asyncio loop object

        Mikrotik duplicate prompt in connection, so we should use pattern like
        prompt .* prompt.
        For disabling colors in CLI output we should user this username = username+c
        '+c' disables colors
        '+t' disable auto term capabilities detection
        '+200w' set terminal width to 200 rows
        """
        super().__init__(*args, **kwargs)
        self._base_pattern = r"\[.*?\] \>.*\[.*?\] \>"
        self._username += "+ct200w"
        self._ansi_escape_codes = True

    _pattern = r"\[.*?\] (\/.*?)?\>"

    async def _flush_buffer(self):
        await self._conn._read_until_prompt()

    async def _set_base_prompt(self):
        """
        Setting two important vars
        * base_prompt - textual prompt in CLI (usually hostname)
        * base_pattern - regexp for finding the end of command. IT's platform specific parameter

        For Mikrotik devices base_pattern is "r"\[.*?\] (\/.*?)?\>"
        """
        self._logger.info("Setting base prompt")
        self._base_pattern = type(self)._pattern
        prompt = await self._find_prompt()
        user = ""
        # Strip off trailing terminator
        prompt = prompt[1:-3]
        if "@" in prompt:
            prompt = prompt.split("@")[1]
        self._base_prompt = prompt
        self._logger.debug("Base Prompt: %s" % self._base_prompt)
        self._logger.debug("Base Pattern: %s" % self._base_pattern)
        return self._base_prompt

    async def _find_prompt(self):
        """Finds the current network device prompt, last line only."""
        self._logger.info("Finding prompt")
        prompt = await self.send_command_expect("\r")
        prompt = prompt.strip()
        if self._ansi_escape_codes:
            prompt = self._strip_ansi_escape_codes(prompt)
        if not prompt:
            raise ValueError("Unable to find prompt: %s" % prompt)
        self._logger.debug("Prompt: %s" % prompt)
        return prompt

    @staticmethod
    def _normalize_cmd(command):
        """Specific trailing newline for Mikrotik"""
        command = command.rstrip("\n")
        command += "\r"
        return command



import setuptools
from aionet import __version__, __author__

with open("README.rst", "r") as fh:
  long_description = fh.read()

setuptools.setup(
  name="aionet",
  version=__version__,
  author=__author__,
  author_email="aaqrabaw@gmail.com",
  description="async networking SDK",
  long_description=long_description,
  url="https://github.com/Ali-aqrabawi/aionet",
  packages=setuptools.find_packages(),
  include_package_data=True,
  install_requires=['asyncssh', 'textfsm'],
  classifiers=[
    "Development Status :: 4 - Beta",
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
  ],
)



"""
Base Device
"""

import asyncio
import re

# from aionet.logger import logger
from aionet.logging import logger, aionetLoggerAdapter
from aionet.version import __version__
from aionet.exceptions import AionetConnectionError
from aionet import utils
from aionet.connections import SSHConnection, TelnetConnection


class BaseDevice(object):

    def __init__(
            self,
            ip=u"",
            username=u"",
            password=u"",
            port=None,
            protocol='ssh',
            device_type=u"",
            timeout=15,
            loop=None,
            known_hosts=None,
            local_addr=None,
            client_keys=None,
            passphrase=None,
            tunnel=None,
            pattern=None,
            agent_forwarding=False,
            agent_path=(),
            client_version=u"aionet-" + __version__,
            family=0,
            kex_algs=(),
            encryption_algs=(),
            mac_algs=(),
            compression_algs=(),
            signature_algs=(),
    ):
        """
        Initialize base class for asynchronous working with network devices

        :param ip: ip address for connection
        :param username: username for logging to device
        :param password: user password for logging to device
        :param port: port number. Default is 22 for ssh and 23 for telnet
        :param protocol: connection protocol (telnet or ssh)
        :param device_type: network device type
        :param timeout: timeout in second for getting information from channel
        :param loop: asyncio loop object
        :param known_hosts: file with known hosts. Default is None (no policy). With () it will use default file
        :param local_addr: local address for binding source of tcp connection
        :param client_keys: path for client keys. Default in None. With () it will use default file in OS
        :param passphrase: password for encrypted client keys
        :param tunnel: An existing SSH connection that this new connection should be tunneled over
        :param pattern: pattern for searching the end of device prompt.
                Example: r"{hostname}.*?(\(.*?\))?[{delimeters}]"
        :param agent_forwarding: Allow or not allow agent forward for server
        :param agent_path:
            The path of a UNIX domain socket to use to contact an ssh-agent
            process which will perform the operations needed for client
            public key authentication. If this is not specified and the environment
            variable `SSH_AUTH_SOCK` is set, its value will be used as the path.
            If `client_keys` is specified or this argument is explicitly set to `None`,
            an ssh-agent will not be used.
        :param client_version: version which advertised to ssh server
        :param family:
           The address family to use when creating the socket. By default,
           the address family is automatically selected based on the host.
        :param kex_algs:
            A list of allowed key exchange algorithms in the SSH handshake,
            taken from `key exchange algorithms
            <https://asyncssh.readthedocs.io/en/latest/api.html#kexalgs>`_
        :param encryption_algs:
            A list of encryption algorithms to use during the SSH handshake,
            taken from `encryption algorithms
            <https://asyncssh.readthedocs.io/en/latest/api.html#encryptionalgs>`_
        :param mac_algs:
            A list of MAC algorithms to use during the SSH handshake, taken
            from `MAC algorithms <https://asyncssh.readthedocs.io/en/latest/api.html#macalgs>`_
        :param compression_algs:
            A list of compression algorithms to use during the SSH handshake,
            taken from `compression algorithms
            <https://asyncssh.readthedocs.io/en/latest/api.html#compressionalgs>`_, or
            `None` to disable compression
        :param signature_algs:
            A list of public key signature algorithms to use during the SSH
            handshake, taken from `signature algorithms
            <https://asyncssh.readthedocs.io/en/latest/api.html#signaturealgs>`_
        

        :type host: str
        :type username: str
        :type password: str
        :type port: int
        :type protocol: str
        :type device_type: str
        :type timeout: int
        :type known_hosts:
            *see* `SpecifyingKnownHosts
            <https://asyncssh.readthedocs.io/en/latest/api.html#specifyingknownhosts>`_
        :type loop: :class:`AbstractEventLoop <asyncio.AbstractEventLoop>`
        :type pattern: str
        :type tunnel: :class:`BaseDevice <aionet.vendors.BaseDevice>`
        :type family:
            :class:`socket.AF_UNSPEC` or :class:`socket.AF_INET` or :class:`socket.AF_INET6`
        :type local_addr: tuple(str, int)
        :type client_keys:
            *see* `SpecifyingPrivateKeys
            <https://asyncssh.readthedocs.io/en/latest/api.html#specifyingprivatekeys>`_
        :type passphrase: str
        :type agent_path: str
        :type agent_forwarding: bool
        :type client_version: str
        :type kex_algs: list[str]
        :type encryption_algs: list[str]
        :type mac_algs: list[str]
        :type compression_algs: list[str]
        :type signature_algs: list[str]
        """
        if ip:
            self.host = ip
        else:
            raise ValueError("Host must be set")

        self._device_type = device_type
        self._timeout = timeout
        self._protocol = protocol
        if loop is None:
            self._loop = asyncio.get_event_loop()
        else:
            self._loop = loop

        if self._protocol == 'ssh':
            self._port = port or 22
            self._port = int(self._port)
            self._ssh_connect_params_dict = {
                "host": self.host,
                "port": self._port,
                "username": username,
                "password": password,
                "known_hosts": known_hosts,
                "local_addr": local_addr,
                "client_keys": client_keys,
                "passphrase": passphrase,
                "tunnel": tunnel,
                "agent_forwarding": agent_forwarding,
                "loop": loop,
                "family": family,
                "agent_path": agent_path,
                "client_version": client_version,
                "kex_algs": kex_algs,
                "encryption_algs": encryption_algs,
                "mac_algs": mac_algs,
                "compression_algs": compression_algs,
                "signature_algs": signature_algs,
            }
        elif self._protocol == 'telnet':
            self._port = port or 23
            self._port = int(self._port)
            self._telnet_connect_params_dict = {
                "host": self.host,
                "port": self._port,
                "username": username,
                "password": password,
            }
        else:
            raise ValueError("unknown protocol {} , only telnet and ssh supported".format(self._protocol))
        self.current_terminal = None

        if pattern is not None:
            self._pattern = pattern

        self._ansi_escape_codes = False

        self._logger = aionetLoggerAdapter(logger, extra={'host': self.host})
        self._logger._host = self.host

    _delimiter_list = [">", "#"]
    """All this characters will stop reading from buffer. It mean the end of device prompt"""

    _pattern = r"{prompt}.*?(\(.*?\))?[{delimiters}]"
    """Pattern for using in reading buffer. When it found processing ends"""

    async def __aenter__(self):
        """Async Context Manager"""
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Async Context Manager"""
        await self.disconnect()

    async def connect(self):
        """
        Basic asynchronous connection method

        It connects to device and makes some preparation steps for working.
        Usual using 3 functions:

        * _establish_connection() for connecting to device
        * _set_base_prompt() for finding and setting device prompt
        * _disable_paging() for non interactive output in commands
        """
        self._logger.info("Trying to connect to the device")
        try:
            await self._establish_connection()
        except OSError as e:
            raise AionetConnectionError(self.host, None, str(e))
        await self._session_preparation()
        logger.info("Has connected to the device")

    async def _establish_connection(self):
        """Establishing SSH connection to the network device"""
        self._logger.info("Establishing connection")
        # initiate SSH connection
        if self._protocol == 'ssh':
            conn = SSHConnection(**self._ssh_connect_params_dict)
        elif self._protocol == 'telnet':
            conn = TelnetConnection(**self._telnet_connect_params_dict)
        else:
            raise ValueError("only SSH connection is supported")

        await conn.connect()
        self._conn = conn
        self._logger.info("Connection is established")

    async def _session_preparation(self):
        """ Prepare session before start using it """
        await self._flush_buffer()
        await self._set_base_prompt()

    async def _flush_buffer(self):
        """ flush unnecessary data """
        self._logger.debug("Flushing buffers")

        delimiters = map(re.escape, type(self)._delimiter_list)
        delimiters = r"|".join(delimiters)
        # await self.send_new_line(pattern=delimiters)
        await self._conn.read_until_pattern(delimiters)

    async def _set_base_prompt(self):
        """
        Setting two important vars:

            base_prompt - textual prompt in CLI (usually hostname)
            base_pattern - regexp for finding the end of command. It's platform specific parameter

        For Cisco devices base_pattern is "prompt(\(.*?\))?[#|>]
        """
        self._logger.info("Setting base prompt")
        prompt = await self._find_prompt()

        # Strip off trailing terminator
        base_prompt = prompt[:-1]
        if not base_prompt:
            raise ValueError("unable to find base_prompt")
        self._conn.set_base_prompt(base_prompt)

        delimiters = map(re.escape, type(self)._delimiter_list)
        delimiters = r"|".join(delimiters)
        base_prompt = re.escape(base_prompt[:12])
        pattern = type(self)._pattern
        base_pattern = pattern.format(prompt=base_prompt, delimiters=delimiters)
        self._logger.debug("Base Prompt: %s" % base_prompt)
        self._logger.debug("Base Pattern: %s" % base_pattern)
        if not base_pattern:
            raise ValueError("unable to find base_pattern")
        self._conn.set_base_pattern(base_pattern)

    async def _find_prompt(self):
        """Finds the current network device prompt, last line only"""
        self._logger.info("Finding prompt")
        await self.send_new_line(dont_read=True)
        delimiters = map(re.escape, type(self)._delimiter_list)
        delimiters = r"|".join(delimiters)
        prompt = await self._conn.read_until_pattern(delimiters)
        prompt = prompt.strip()
        if self._ansi_escape_codes:
            prompt = self._strip_ansi_escape_codes(prompt)
        if not prompt:
            raise ValueError(
                "Host %s: Unable to find prompt: %s" % (self.host, repr(prompt))
            )
        self._logger.debug("Found Prompt: %s" % repr(prompt))
        return prompt

    async def send_command_timing(self,
                                  command_string,
                                  read_for_seconds=2):
        """
        send command and keep reading for the specified time in wait or until_prompt
        :param command_string: command
        :type command_string: str
        :param read_for_seconds: seconds of reading
        :type read_for_seconds: int
        :return: command output
        """

        output = await self.send_command_expect(command_string, read_for=read_for_seconds)
        return output

    async def send_command(
            self,
            command_string,
            pattern="",
            re_flags=0,
            strip_command=True,
            strip_prompt=True,
            use_textfsm=False
    ):
        """
        Sending command to device (support interactive commands with pattern)

        :param str command_string: command for executing basically in privilege mode
        :param str pattern: pattern for waiting in output (for interactive commands)
        :param re.flags re_flags: re flags for pattern
        :param bool strip_command: True or False for stripping command from output
        :param bool strip_prompt: True or False for stripping ending device prompt
        :param use_textfsm: True or False for parsing output with textfsm templates
                            download templates from https://github.com/networktocode/ntc-templates
                            and set  NET_TEXTFSM environment to pint to ./ntc-templates/templates
        :return: The output of the command
        """
        self._logger.info("Sending command")

        command_string = self._normalize_cmd(command_string)
        self._logger.debug(
            "Send command: %s" % repr(command_string)
        )

        output = await self.send_command_expect(command_string, pattern, re_flags)

        # Some platforms have ansi_escape codes
        if self._ansi_escape_codes:
            output = self._strip_ansi_escape_codes(output)
        output = self._normalize_linefeeds(output)
        if strip_prompt:
            output = self._strip_prompt(output)
        if strip_command:
            output = self._strip_command(command_string, output)

        if use_textfsm:
            self._logger.info("parsing output using texfsm, command=%r," % command_string)
            output = utils.get_structured_data(output, self._device_type, command_string)

        logger.debug(
            "Host %s: Send command output: %s" % (self.host, repr(output))
        )
        return output

    def _strip_prompt(self, a_string):
        """Strip the trailing router prompt from the output"""
        self._logger.info("Stripping prompt")
        response_list = a_string.split("\n")
        last_line = response_list[-1]
        if self._conn._base_prompt in last_line:
            return "\n".join(response_list[:-1])
        else:
            return a_string

    @staticmethod
    def _strip_backspaces(output):
        """Strip any backspace characters out of the output"""
        backspace_char = "\x08"
        return output.replace(backspace_char, "")

    @staticmethod
    def _strip_command(command_string, output):
        """
        Strip command_string from output string

        Cisco IOS adds backspaces into output for long commands (i.e. for commands that line wrap)
        """
        backspace_char = "\x08"

        # Check for line wrap (remove backspaces)
        if backspace_char in output:
            output = output.replace(backspace_char, "")
            output_lines = output.split("\n")
            new_output = output_lines[1:]
            return "\n".join(new_output)
        else:
            command_length = len(command_string)
            return output[command_length:]

    @staticmethod
    def _normalize_linefeeds(a_string):
        """Convert '\r\r\n','\r\n', '\n\r' to '\n"""
        newline = re.compile(r"(\r\r\n|\r\n|\n\r)")
        return newline.sub("\n", a_string)

    @staticmethod
    def _normalize_cmd(command):
        """Normalize CLI commands to have a single trailing newline"""
        command = command.rstrip("\n")
        command += "\n"
        return command

    async def send_new_line(self, pattern='', dont_read=False):
        """ Sending new line """
        return await self.send_command_expect('\n', pattern=pattern, dont_read=dont_read)

    async def send_command_expect(self, command,
                                  pattern='',
                                  re_flags=0, dont_read=False,
                                  read_for=0):
        """ Send a single line of command and readuntil prompte"""
        self._conn.send(self._normalize_cmd(command))
        if dont_read:
            return ''
        if pattern:
            output = await self._conn.read_until_prompt_or_pattern(pattern, re_flags, read_for=read_for)

        else:
            output = await self._conn.read_until_prompt(read_for=read_for)

        return output

    async def send_config_set(self, config_commands=None):
        """
        Sending configuration commands to device

        The commands will be executed one after the other.

        :param list config_commands: iterable string list with commands for applying to network device
        :return: The output of this commands
        """
        self._logger.info("Sending configuration settings")
        assert isinstance(config_commands, list), "config_commands must be list not %s" % type(config_commands)
        if config_commands is None:
            return ""

        # Send config commands
        self._logger.debug("Config commands: %s" % config_commands)
        config_commands = ['\n'] + config_commands  # to get ride of unwanted messages.
        output = ""
        for cmd in config_commands:
            output += await self.send_command_expect(cmd)

        if self._ansi_escape_codes:
            output = self._strip_ansi_escape_codes(output)

        output = self._normalize_linefeeds(output)
        self._logger.debug(
            "Config commands output: %s" % repr(output)
        )
        return output

    @staticmethod
    def _strip_ansi_escape_codes(string_buffer):
        return utils.strip_ansi_escape_codes(string_buffer)

    async def disconnect(self):
        """ Gracefully close the SSH connection """
        self._logger.info("Disconnecting")
        await self._conn.close()



from aionet.vendors.devices.arista import AristaEOS
from aionet.vendors.devices.aruba import ArubaAOS8, ArubaAOS6
from aionet.vendors.devices.base import BaseDevice
from aionet.vendors.devices.cisco import CiscoNXOS, CiscoIOSXR, CiscoASA, CiscoIOS
from aionet.vendors.devices.base_comware import BaseComwareDevice
from aionet.vendors.devices.fujitsu import FujitsuSwitch
from aionet.vendors.devices.hp import HPComware, HPComwareLimited
from aionet.vendors.devices.base_ios import BaseIOSDevice
from aionet.vendors.devices.juniper import JuniperJunOS
from aionet.vendors.devices.base_junos import BAseJunOSDevice
from aionet.vendors.devices.mikrotik import MikrotikRouterOS
from aionet.vendors.devices.terminal import Terminal
from aionet.vendors.devices.ubiquiti import UbiquityEdgeSwitch

__all__ = (
    "CiscoASA",
    "CiscoIOS",
    "CiscoIOSXR",
    "CiscoNXOS",
    "HPComware",
    "HPComwareLimited",
    "FujitsuSwitch",
    "MikrotikRouterOS",
    "JuniperJunOS",
    "BAseJunOSDevice",
    "AristaEOS",
    "ArubaAOS6",
    "ArubaAOS8",
    "BaseDevice",
    "BaseIOSDevice",
    "BaseComwareDevice",
    "Terminal",
    "arista",
    "aruba",
    "cisco",
    "fujitsu",
    "hp",
    "juniper",
    "mikrotik",
    "UbiquityEdgeSwitch",
)
